from nopasaran.interpreters.interpreter import Interpreter
from nopasaran.primitives.action_primitives.action_primitives import ActionPrimitives


class ActionInterpreter(Interpreter):
    """
    A custom ActionInterpreter class that inherits from Interpreter.
    This class interprets actions in the context of a command-line interface.
    """

    @classmethod
    def evaluate(cls, line, state_variables):
        """
        Evaluate the action line using the provided state variables.

        Args:
            line (str): The action line to evaluate.
            state_variables: The state variables to use during evaluation.

        Returns:
            The result of the evaluation.
        """
        action_primitives = ActionPrimitives()
        super().evaluate(line, state_variables, action_primitives)



# checks.py
# This file contains the functions that are used to verify the frames
# Each function takes an event and a set of parameters and returns a boolean
# If the test does not apply to the event, the function should return None
# If the test passes, the function should return True
# If the test fails, the function should return False

def check_header_field(event, name, value=None):
    """
    Check if a header field exists and optionally check its value.
    
    Args:
        event: The event containing headers
        name: The header name to check for
        value: Optional value to match against. If None, only checks header presence
    
    Returns:
        - True if:
            * value is None and header is found
            * value is provided and matches the header value
        - False if:
            * header is not found
            * value is provided but doesn't match
        - None if the event doesn't have headers
    """
    if hasattr(event, 'headers'):
        for header_name, header_value in event.headers:
            if header_name == name:
                if value is None:
                    return True
                return header_value == value
        return False
    return None

function_map = {
    'check_header_field': check_header_field
}


from nopasaran.decorators import parsing_decorator
from nopasaran.tools.http_2_socket_server import HTTP2SocketServer

class HTTP2ServerPrimitives:
    """
    Class containing HTTP/2 server action primitives for the state machine.
    """

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def create_http_2_server(inputs, outputs, state_machine):
        """
        Create an instance of HTTP2SocketServer.

        Number of input arguments: 2
            - The host
            - The port

        Number of output arguments: 1
            - The created HTTP2SocketServer instance

        Args:
            inputs (List[str]): The list of input variable names. No input arguments for this method.

            outputs (List[str]): The list of output variable names. It contains one output argument:
                - The name of the variable to store the HTTP2SocketServer instance.

            state_machine: The state machine object.

        Returns:
            None
        """
        host = state_machine.get_variable_value(inputs[0])
        port = state_machine.get_variable_value(inputs[1])
        port = int(port)
        server = HTTP2SocketServer(host, port)
        state_machine.set_variable_value(outputs[0], server)


    @staticmethod
    @parsing_decorator(input_args=4, output_args=2)
    def start_http_2_server(inputs, outputs, state_machine):
        """
        Start the HTTP/2 server.

        Number of input arguments: 4
            - The HTTP2SocketServer instance
            - The tls_enabled flag
            - The connection settings for the server
            - The cloudflare_origin flag
        Number of output arguments: 2
            - The event name
            - The message

        Args:
            inputs (List[str]): The list of input variable names containing:
                - The name of the HTTP2SocketServer instance variable
                - The name of the tls_enabled flag variable
                - The name of the connection settings variable
                - The name of the cloudflare_origin flag variable

            outputs (List[str]): The list of output variable names. It contains two output arguments:
                - The name of the variable to store the event name
                - The name of the variable to store the message

            state_machine: The state machine object.

        Returns:
            None

        Possible events:
            - EventNames.ERROR
            - EventNames.SERVER_STARTED
            - EventNames.TIMEOUT
        """
        server = state_machine.get_variable_value(inputs[0])
        tls_enabled = state_machine.get_variable_value(inputs[1])
        connection_settings_server = state_machine.get_variable_value(inputs[2])
        cloudflare_origin = state_machine.get_variable_value(inputs[3])

        event, msg = server.start(tls_enabled, connection_settings_server, cloudflare_origin)
        state_machine.set_variable_value(outputs[0], event)
        state_machine.set_variable_value(outputs[1], msg)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=3)
    def wait_for_client_preface(inputs, outputs, state_machine):
        """
        Wait for the client's connection preface.

        Number of input arguments: 1
            - The HTTP2SocketServer instance

        Number of output arguments: 3
            - The event name
            - The message
            - The frames received

        Args:
            inputs (List[str]): The list of input variable names containing:
                - The name of the HTTP2SocketServer instance variable

            outputs (List[str]): The list of output variable names. It contains two output arguments:
                - The name of the variable to store the event name
                - The name of the variable to store the message
                - The name of the variable to store the frames received
                
            state_machine: The state machine object.

        Returns:
            None
        """
        server = state_machine.get_variable_value(inputs[0])
        event, msg, frames = server.wait_for_preface()
        state_machine.set_variable_value(outputs[0], event)
        state_machine.set_variable_value(outputs[1], msg)
        state_machine.set_variable_value(outputs[2], frames)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=2)
    def wait_for_client_ack(inputs, outputs, state_machine):
        """
        Wait for the client's SETTINGS_ACK frame.

        Number of input arguments: 1
            - The HTTP2SocketServer instance

        Number of output arguments: 2
            - The event name
            - The message

        Args:
            inputs (List[str]): The list of input variable names containing:
                - The name of the HTTP2SocketServer instance variable

            outputs (List[str]): The list of output variable names. It contains two output arguments:
                - The name of the variable to store the event name
                - The name of the variable to store the message

            state_machine: The state machine object.

        Returns:
            None
        """
        server = state_machine.get_variable_value(inputs[0])
        event, msg = server.wait_for_preface_ack()
        state_machine.set_variable_value(outputs[0], event)
        state_machine.set_variable_value(outputs[1], msg)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=3)
    def receive_client_frames(inputs, outputs, state_machine):
        """
        Receive the client's frames and handle the tests.

        Number of input arguments: 2
            - The HTTP2SocketServer instance
            - The client frames to receive

        Number of output arguments: 3
            - The event name
            - The message
            - The frames received

        Args:
            inputs (List[str]): The list of input variable names containing:
                - The name of the HTTP2SocketServer instance variable
                - The name of the client frames variable

            outputs (List[str]): The list of output variable names. It contains two output arguments:
                - The name of the variable to store the event name
                - The name of the variable to store the message
                - The name of the variable to store the frames received

            state_machine: The state machine object.

        Returns:
            None

        Possible events:
            - EventNames.TIMEOUT
            - EventNames.RESET_RECEIVED
            - EventNames.GOAWAY_RECEIVED
            - EventNames.REJECTED
            - EventNames.RECEIVED_FRAMES
        """
        server = state_machine.get_variable_value(inputs[0])
        test_frames = state_machine.get_variable_value(inputs[1])
        event, msg, frames_received = server.receive_test_frames(test_frames)
        state_machine.set_variable_value(outputs[0], event)
        state_machine.set_variable_value(outputs[1], msg)
        state_machine.set_variable_value(outputs[2], frames_received)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=3)
    def send_server_frames(inputs, outputs, state_machine):
        """
        Send the server's frames.

        Number of input arguments: 2
            - The HTTP2SocketServer instance
            - The frames to send

        Number of output arguments: 3
            - The event name
            - The frames sent
            - The message

        Args:
            inputs (List[str]): The list of input variable names containing:
                - The name of the HTTP2SocketServer instance variable
                - The name of the server frames variable

            outputs (List[str]): The list of output variable names. It contains two output arguments:
                - The name of the variable to store the event name
                - The name of the variable to store the frames sent
                - The name of the variable to store the message

            state_machine: The state machine object.

        Returns:
            None

        Possible events:
            - EventNames.FRAMES_SENT
            - EventNames.GOAWAY_RECEIVED
            - EventNames.RESET_RECEIVED
            - EventNames.REJECTED
        """
        server = state_machine.get_variable_value(inputs[0])
        server_frames = state_machine.get_variable_value(inputs[1])
        event, frames_sent, msg = server.send_frames(server_frames)
        state_machine.set_variable_value(outputs[0], event)
        state_machine.set_variable_value(outputs[1], frames_sent)
        state_machine.set_variable_value(outputs[2], msg)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def close_http_2_server(inputs, outputs, state_machine):
        """
        Close the HTTP/2 server connection gracefully.
        
        Number of input arguments: 1
            - The HTTP2SocketServer instance
            
        Number of output arguments: 1
            - The event name
        """
        server = state_machine.get_variable_value(inputs[0])
        event = server.close()
        state_machine.set_variable_value(outputs[0], event)
        state_machine.trigger_event(event)


from enum import Enum


class Command(Enum):
    """
    Enum representing commands.
    
    This enum represents different commands used in state transitions.
    """

    EXECUTE_ACTION = 0
    ASSIGN_VARIABLES = 1
    SET_STATE = 2



import time
import select
import h2.connection
import h2.config
import h2.events
from nopasaran.tools.checks import function_map
from nopasaran.definitions.events import EventNames
from nopasaran.http_2_utils import (
    SSL_CONFIG,
    send_frame
)

class HTTP2SocketBase:
    def __init__(self, host: str, port: int):
        self.host = host
        self.port = port
        self.sock = None
        self.conn = None
        self.MAX_RETRY_ATTEMPTS = 3
        self.TIMEOUT = 5.0
        self.cloudflare_origin = False

    def _receive_frame(self, timeout = None) -> bytes:
        """Helper method to receive data"""
        try:
            start_time = time.time()
            socket_to_check = self.sock if not hasattr(self, 'client_socket') else self.client_socket
            
            if timeout is None:
                timeout = self.TIMEOUT
            
            while True:
                elapsed_time = time.time() - start_time
                if elapsed_time > timeout:
                    return None
                
                ready_to_read, _, _ = select.select([socket_to_check], [], [], timeout)
                
                if ready_to_read:
                    frame = socket_to_check.recv(SSL_CONFIG.MAX_BUFFER_SIZE)
                    return frame
        except (ConnectionResetError, BrokenPipeError) as e:
            # Connection was terminated by the peer
            return None
        except Exception as e:
            # Log unexpected errors
            return None

    def send_frames(self, frames):
        """Send frames with GOAWAY checking after each frame"""
        socket_to_use = self.sock if not hasattr(self, 'client_socket') else self.client_socket
        sent_frames = []
        is_server = hasattr(self, 'client_socket')
        
        for frame in frames:
            # Send the current frame
            send_frame(self.conn, socket_to_use, frame, is_server, self.cloudflare_origin)
            sent_frames.append(frame)

            # Flush socket after each frame to ensure it's sent
            outbound_data = self.conn.data_to_send()
            if outbound_data:
                socket_to_use.sendall(outbound_data)
            
            # Use non-blocking check after each frame with minimal timeout
            ready_to_read, _, _ = select.select([socket_to_use], [], [], 0.01)  # 10ms timeout
            
            if ready_to_read:
                # Data is available, check for GOAWAY
                data = self._receive_frame(timeout=0.05)  # Very short timeout
                if data:
                    events = self.conn.receive_data(data)
                    for event in events:
                        if isinstance(event, h2.events.ConnectionTerminated):
                            return (
                                EventNames.GOAWAY_RECEIVED.name,
                                str(sent_frames),
                                f"Connection terminated by peer: Received GOAWAY frame after sending {len(sent_frames)} of {len(frames)} frames. Error code {event.error_code}. Additional data: {event.additional_data}."
                            )
                        elif isinstance(event, h2.events.StreamReset):
                            return (
                                EventNames.RESET_RECEIVED.name,
                                str(sent_frames),
                                f"Stream {event.stream_id} reset by peer: Received StreamReset frame after sending frame #{len(sent_frames)} of {len(frames)}: {frame.get('type')}. Error code {event.error_code}."
                            )
                        elif isinstance(event, h2.events.ResponseReceived) or isinstance(event, h2.events.RequestReceived):
                            # check for 5xx status code
                            for header_name, header_value in event.headers:
                                if header_name == ':status' and int(header_value) >= 500:
                                        return (
                                            EventNames.REJECTED.name,
                                            str(sent_frames),
                                            f"Received 5xx status code {header_value} after receiving {len(sent_frames)}/{len(frames)} frames."
                                        )

        # Final check for GOAWAY with a slightly longer timeout
        data = self._receive_frame(0.1)
        if data is not None:
            events = self.conn.receive_data(data)
            for event in events:
                if isinstance(event, h2.events.ConnectionTerminated):
                    return (
                        EventNames.GOAWAY_RECEIVED.name,
                        str(sent_frames),
                        f"Connection terminated by peer: Received GOAWAY frame with error code {event.error_code}. Additional data: {event.additional_data}."
                    )
        
        return EventNames.FRAMES_SENT.name, str(sent_frames), f"Successfully sent {len(sent_frames)} frames."

    def _handle_test(self, event, frame) -> None | str:
        """
        Handle test cases for received frames.
        Tests contain a single check with dictionary parameters and conditional results.

        Returns:
            - The result of the test if it passed
            - None if no tests were found for that frame
        """
        test = frame.get('test', {})
        
        function_name = test.get('function')
        params = test.get('params', {})
        
        function = function_map.get(function_name)

        # Execute the function with unpacked dictionary parameters
        result = function(event, **params)
        
        # Return based on the test result and specified conditions
        if result == True:
            return test['if_true']
        elif result == False:
            return test['if_false']
        elif result is None:
            return None
    
    def receive_test_frames(self, test_frames) -> str:
        """Wait for test frames with adaptive timeout"""
        frames_received = []
        expected_frame_count = len(test_frames)
        last_frame_time = time.time()
        start_time = time.time()

        
        while len(frames_received) < expected_frame_count:
            # Check for overall timeout
            if time.time() - start_time > self.TIMEOUT * 2:  # Double timeout for overall operation
                return EventNames.TIMEOUT.name, f"Overall timeout after {self.TIMEOUT * 2}s. Received {len(frames_received)} of {expected_frame_count} expected frames.", str(frames_received)
            
            # Check for timeout since last frame
            if time.time() - last_frame_time > self.TIMEOUT:
                return EventNames.TIMEOUT.name, f"Timeout after {self.TIMEOUT}s since last frame. Received {len(frames_received)} of {expected_frame_count} expected frames.", str(frames_received)
            
            data = self._receive_frame()
            
            if data is not None:
                last_frame_time = time.time()
                events = self.conn.receive_data(data)
                
                for event in events:
                    if isinstance(event, h2.events.StreamReset):
                        if event.stream_id == 1:
                            continue
                        return EventNames.RESET_RECEIVED.name, f"Stream {event.stream_id} reset after receiving {len(frames_received)}/{expected_frame_count} frames. Got error code {event.error_code}.", str(events)

                    if isinstance(event, h2.events.ConnectionTerminated):
                        return EventNames.GOAWAY_RECEIVED.name, f"Peer terminated connection after receiving {len(frames_received)}/{expected_frame_count} frames. Got error code {event.error_code}.", str(events)
                    
                    # Skip initial settings frame
                    if isinstance(event, h2.events.RemoteSettingsChanged):
                        settings = event.changed_settings.items()
                        if len(settings) >= 2:
                            continue
                    
                    # Skip initial settings ACK
                    if isinstance(event, h2.events.SettingsAcknowledged) \
                        or isinstance(event, h2.events.StreamEnded) \
                        or isinstance(event, h2.events.WindowUpdated):
                        continue

                    if isinstance(event, h2.events.DataReceived):
                        # Compare bytes to bytes
                        if event.data == b'Connection test from client':
                            continue

                    if isinstance(event, h2.events.ResponseReceived) or isinstance(event, h2.events.RequestReceived):
                        # check for 5xx status code
                        for header_name, header_value in event.headers:
                            if header_name == ':status' and int(header_value) >= 500:
                                return EventNames.REJECTED.name, f"Received 5xx status code {header_value} after receiving {len(frames_received)}/{expected_frame_count} frames.", str(event)

                    # Filter for connection-test headers
                    if isinstance(event, h2.events.RequestReceived):
                        for header_name, header_value in event.headers:
                            if header_name == ':path' and header_value == '/connection-test':
                                continue

                    frames_received.append(event)

                    # Move test handling before the frame count check
                    for frame in test_frames:
                        if frame.get('test'):
                            result = self._handle_test(event, frame)
                            if result is not None:
                                return EventNames.RECEIVED_FRAMES.name, f"Test result: {result}", str(frames_received)
                    
                    if len(frames_received) == expected_frame_count:
                        return EventNames.RECEIVED_FRAMES.name, f"Successfully received all {len(frames_received)}/{expected_frame_count} frames.", str(frames_received)

    def close(self):
        """Close the HTTP/2 connection and clean up resources"""
        try:
            if self.conn:
                # Send GOAWAY frame to indicate graceful shutdown
                self.conn.close_connection()
                socket_to_use = self.sock if not hasattr(self, 'client_socket') else self.client_socket
                if socket_to_use:
                    socket_to_use.sendall(self.conn.data_to_send())
            
            # Close sockets
            if hasattr(self, 'client_socket') and self.client_socket:
                self.client_socket.close()
            if self.sock:
                self.sock.close()
            
            # Clear references
            self.conn = None
            if hasattr(self, 'client_socket'):
                self.client_socket = None
            self.sock = None
            
            return EventNames.CONNECTION_CLOSED.name
        except Exception as e:
            return EventNames.ERROR.name


import json

from nopasaran.decorators import parsing_decorator

class NestedMachinePrimitives:
    """
    Class containing primitives for interacting with nested state machines.
    """

    @staticmethod
    @parsing_decorator(input_args=1, output_args=0, optional_inputs=True, optional_outputs=True)
    def call(inputs, outputs, state_machine):
        """
        Load a nested machine identified by the given name and call its start method with the provided parameters.
        The mandatory input argument is the name of the nested state machine to call.
        The optional input arguments are the parameters assigned to the nested state machine.
        The optional output arguments are the values returned by the nested state machine and will be stored in the
        specified optional output variables. The number of optional output arguments should match the number of optional
        input arguments specified in the 'return_values' primitive used within the nested state machine.

        Number of input arguments: 1

        Number of output arguments: 0

        Optional input arguments: Yes

        Optional output arguments: Yes

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the nested state machine to call, and optional input arguments representing the parameters assigned to the nested state machine.
            
            outputs (List[str]): The list of output variable names. It contains the names of the variables where the returned values will be stored (optional).
            
            state_machine: The state machine object.

        Returns:
            None
        """
        nested_state_json = json.load(open('.'.join((inputs[0], 'json'))))
        parameters = []
        for nested_variables in inputs[1:]:
            parameters.append(state_machine.get_variable_value(nested_variables))
        nested_machine = state_machine.get_nested_machine(nested_state_json, parameters)
        nested_machine.start()
        for index in range(len(nested_machine.returned)):
            state_machine.set_variable_value(outputs[index], nested_machine.get_variable_value(nested_machine.returned[index]))

    @staticmethod
    @parsing_decorator(input_args=0, output_args=0, optional_outputs=True)
    def get_parameters(inputs, outputs, state_machine):
        """
        Retrieve the parameters of the current nested state machine that were passed from the parent state machine
        using the 'call' primitive. These parameters are optional input arguments specified in the 'call' action of
        the parent state machine. The retrieved parameter values will be stored in the specified optional output
        arguments for further use within the nested machine. The number of optional input arguments in the 'call'
        primitive and the number of output arguments in 'get_parameters' should match.

        Number of input arguments: 0

        Number of output arguments: 0

        Optional input arguments: No

        Optional output arguments: Yes

        Args:
            inputs (List[str]): The list of input variable names.
            
            outputs (List[str]): The list of output variable names. It contains the names of the variables where the parameter values will be stored (optional).
            
            state_machine: The state machine object.

        Returns:
            None
        """
        for index in range(len(state_machine.parameters)):
            state_machine.set_variable_value(outputs[index], state_machine.parameters[index])

    @staticmethod
    @parsing_decorator(input_args=0, output_args=0, optional_inputs=True, optional_outputs=False)
    def return_values(inputs, outputs, state_machine):
        """
        This primitive is used within a nested state machine. It enables you to specify the values that will be returned
        by the nested state machine when it completes its execution. The returned values can be used by the calling state
        machine for further processing or decision-making.

        Number of input arguments: 0

        Number of output arguments: 0

        Optional input arguments: Yes

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names.

            outputs (List[str]): The list of output variable names.
            
            state_machine: The state machine object.

        Returns:
            None
        """
        state_machine.returned = inputs


import time

from scapy.all import send as sendpacket

from nopasaran.definitions.events import EventNames
from nopasaran.decorators import parsing_decorator


class DataChannelPrimitives:
    """
    Class containing data channel action primitives for the state machine.
    """

    @staticmethod
    @parsing_decorator(input_args=1, output_args=0)
    def send(inputs, outputs, state_machine):
        """
        Send the packet stored in the variable with the given name from the machine's state using the machine's
        network interface. Triggers the event PACKET_SENT.

        Number of input arguments: 1

        Number of output arguments: 0

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable storing the packet to be sent.
            
            outputs (List[str]): The list of output variable names.
            
            state_machine: The state machine object.

        Returns:
            None
        """
        sendpacket(state_machine.get_variable_value(inputs[0]))
        state_machine.trigger_event(EventNames.PACKET_SENT.name)

    @staticmethod
    @parsing_decorator(input_args=0, output_args=1)
    def listen(inputs, outputs, state_machine):
        """
        Start the packet sniffer and store the captured packets in a list stored in the machine's state.

        Number of input arguments: 0

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names.
            
            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the captured packets.
            
            state_machine: The state machine object.

        Returns:
            None
        """
        state_machine.start_sniffer()
        state_machine.set_variable_value(outputs[0], [])
        state_machine.update_sniffer_queue(state_machine.get_variable_value(outputs[0]))

    @staticmethod
    @parsing_decorator(input_args=1, output_args=0)
    def packet_filter(inputs, outputs, state_machine):
        """
        Set the packet filter for the packet sniffer.

        Number of input arguments: 1

        Number of output arguments: 0

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the new packet filter value.
            
            outputs (List[str]): The list of output variable names.
            
            state_machine: The state machine object.

        Returns:
            None
        """
        state_machine.update_sniffer_filter(state_machine.get_variable_value(inputs[0]))

    @staticmethod
    @parsing_decorator(input_args=2, output_args=0)
    def wait_packet_signal(inputs, outputs, state_machine):
        """
        Wait for a packet to be available in the sniffer's packet stack stored in the machine's state.
        The sniffer's packet stack is created and populated in the 'listen' primitive.
        If a packet becomes available within the specified timeout (second mandatory input argument),
        triggers the event PACKET_AVAILABLE. Otherwise, triggers the event TIMEOUT.

        Number of input arguments: 2

        Number of output arguments: 0

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments, which are the name of the packet stack variable and the timeout value.
            
            outputs (List[str]): The list of output variable names.
            
            state_machine: The state machine object.

        Returns:
            None
        """
        timeout = False
        start_time = time.time()
        while True:
            stack = state_machine.get_variable_value(inputs[0])
            if len(stack) > 0:
                state_machine.trigger_event(EventNames.PACKET_AVAILABLE.name)
                break
            if time.time() - start_time > float(state_machine.get_variable_value(inputs[1])):
                timeout = True
                break
        if timeout:
            state_machine.trigger_event(EventNames.TIMEOUT.name)


from nopasaran.interpreters.interpreter import Interpreter
from nopasaran.primitives.condition_primitives.condition_primitives import ConditionPrimitives


class ConditionInterpreter(Interpreter):
    """
    A custom ConditionInterpreter class that inherits from Interpreter.
    This class interprets conditions in the context of a command-line interface.
    """

    @classmethod
    def evaluate(cls, line, state_variables):
        """
        Evaluate the condition line using the provided state variables.

        Args:
            line (str): The condition line to evaluate.
            state_variables: The state variables to use during evaluation.

        Returns:
            The result of the evaluation.
        """
        condition_primitives = ConditionPrimitives()
        return super().evaluate(line, state_variables, condition_primitives)



import nopasaran.utils as utils
from nopasaran.decorators import parsing_decorator

class UDPPrimitives:
    """
    Class containing UDP action primitives for the state machine.
    """

    @staticmethod
    @parsing_decorator(input_args=0, output_args=1)
    def create_UDP_packet(inputs, outputs, state_machine):
        """
        Create a UDP packet and store it in an output variable in the machine's state.

        Number of input arguments: 0

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the created UDP packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        state_machine.set_variable_value(outputs[0], utils.create_UDP_packet())


    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_UDP_sport(inputs, outputs, state_machine):
        """
        Set the source port of a UDP packet.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the UDP packet.
                - The name of the variable containing the source port to set.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument,
                which is the name of the variable to store the modified UDP packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        state_machine.set_variable_value(outputs[0], state_machine.get_variable_value(inputs[0]))
        utils.set_UDP_sport(state_machine.get_variable_value(outputs[0]), state_machine.get_variable_value(inputs[1]))


    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_UDP_dport(inputs, outputs, state_machine):
        """
        Set the destination port of a UDP packet.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the UDP packet.
                - The name of the variable containing the destination port to set.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument,
                which is the name of the variable to store the modified UDP packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        state_machine.set_variable_value(outputs[0], state_machine.get_variable_value(inputs[0]))
        utils.set_UDP_dport(state_machine.get_variable_value(outputs[0]), state_machine.get_variable_value(inputs[1]))

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def get_UDP_sport(inputs, outputs, state_machine):
        """
        Get the source port from a UDP packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument:
                - The name of the variable containing the UDP packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument,
                which is the name of the variable to store the source port.

            state_machine: The state machine object.

        Returns:
            None
        """
        udp_packet = state_machine.get_variable_value(inputs[0])
        sport = udp_packet['UDP'].sport

        state_machine.set_variable_value(outputs[0], sport)


    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def get_UDP_dport(inputs, outputs, state_machine):
        """
        Get the destination port from a UDP packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument:
                - The name of the variable containing the UDP packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument,
                which is the name of the variable to store the destination port.

            state_machine: The state machine object.

        Returns:
            None
        """
        udp_packet = state_machine.get_variable_value(inputs[0])
        dport = udp_packet['UDP'].dport

        state_machine.set_variable_value(outputs[0], dport)



from scapy.all import IP, UDP, DNS, DNSQR, DNSRR
from nopasaran.decorators import parsing_decorator


class DNSPrimitives:
    """
    Class containing DNS action primitives for the state machine.
    """

    @staticmethod
    @parsing_decorator(input_args=0, output_args=1)
    def create_DNS_packet(inputs, outputs, state_machine):
        """
        Create a DNS packet and store it in an output variable in the machine's state.

        Number of input arguments: 0

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the created DNS packet.
            
            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = IP()/UDP()/DNS()
        state_machine.set_variable_value(outputs[0], dns_packet)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def disable_DNS_rd_flag(inputs, outputs, state_machine):
        """
        Disable the recursion desired (rd) flag in the DNS packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_packet['DNS'].rd = 0
        state_machine.set_variable_value(outputs[0], dns_packet)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def enable_DNS_rd_flag(inputs, outputs, state_machine):
        """
        Enable the recursion desired (rd) flag in the DNS packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_packet['DNS'].rd = 1
        state_machine.set_variable_value(outputs[0], dns_packet)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def set_DNS_packet_query(inputs, outputs, state_machine):
        """
        Set the DNS packet as a query packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument:
                - The name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_packet['DNS'].qr = 0
        state_machine.set_variable_value(outputs[0], dns_packet)


    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def set_DNS_packet_response(inputs, outputs, state_machine):
        """
        Set the DNS packet as a response packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument:
                - The name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_packet['DNS'].qr = 1
        state_machine.set_variable_value(outputs[0], dns_packet)


    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def get_DNS_transaction_id(inputs, outputs, state_machine):
        """
        Get the transaction ID from the DNS packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the transaction ID.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        transaction_id = dns_packet['DNS'].id
        state_machine.set_variable_value(outputs[0], transaction_id)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_DNS_transaction_id(inputs, outputs, state_machine):
        """
        Set the transaction ID in the DNS packet.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments. The first argument is the name of the variable containing the DNS packet, and the second argument is the name of the variable containing the new transaction ID.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        new_transaction_id = state_machine.get_variable_value(inputs[1])
        dns_packet['DNS'].id = new_transaction_id
        state_machine.set_variable_value(outputs[0], dns_packet)

    @staticmethod
    @parsing_decorator(input_args=0, output_args=1)
    def create_DNS_query(inputs, outputs, state_machine):
        """
        Create a DNS query (DNSQR) and store it in an output variable in the machine's state.

        Number of input arguments: 0

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the created DNS query.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_query = DNSQR()
        state_machine.set_variable_value(outputs[0], dns_query)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def add_DNS_query_to_DNS_packet(inputs, outputs, state_machine):
        """
        Add a DNS query (DNSQR) to a DNS packet.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS packet.
                - The name of the variable containing the DNS query to be added.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_query = state_machine.get_variable_value(inputs[1])

        dns_packet['DNS'].qd = dns_query

        state_machine.set_variable_value(outputs[0], dns_packet)


    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def get_DNS_query_from_DNS_packet(inputs, outputs, state_machine):
        """
        Get the DNS query field (DNSQR) from a DNS packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the DNS query.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_query = dns_packet.qd

        state_machine.set_variable_value(outputs[0], dns_query)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_query_name(inputs, outputs, state_machine):
        """
        Set the domain name (qname) of a DNS query (DNSQR).

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS query.
                - The name of the variable containing the new query name (qname).

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the modified DNS query.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_query = state_machine.get_variable_value(inputs[0])
        new_query_name = state_machine.get_variable_value(inputs[1])
        dns_query.qname = new_query_name
        state_machine.set_variable_value(outputs[0], dns_query)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def get_query_name(inputs, outputs, state_machine):
        """
        Get the domain name (qname) of a DNS query (DNSQR).

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable containing the DNS query.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the qualified domain name.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_query = state_machine.get_variable_value(inputs[0])
        qname = dns_query.qname.decode()

        state_machine.set_variable_value(outputs[0], qname)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_query_type(inputs, outputs, state_machine):
        """
        Set the query type (qtype) of a DNS query (DNSQR).

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS query.
                - The name of the variable containing the new query type.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the modified DNS query.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_query = state_machine.get_variable_value(inputs[0])
        new_query_type = state_machine.get_variable_value(inputs[1])
        dns_query.qtype = new_query_type
        state_machine.set_variable_value(outputs[0], dns_query)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_query_class(inputs, outputs, state_machine):
        """
        Set the query class (qclass) of a DNS query (DNSQR).

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS query.
                - The name of the variable containing the new query class.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the modified DNS query.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_query = state_machine.get_variable_value(inputs[0])
        new_query_class = state_machine.get_variable_value(inputs[1])
        dns_query.qclass = new_query_class
        state_machine.set_variable_value(outputs[0], dns_query)

    @staticmethod
    @parsing_decorator(input_args=0, output_args=1)
    def create_DNS_resource_record(inputs, outputs, state_machine):
        """
        Create a DNS resource record without input arguments and store it in an output variable in the machine's state.

        Number of input arguments: 0

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the created DNS resource record.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_resource_record = DNSRR()
        state_machine.set_variable_value(outputs[0], dns_resource_record)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_DNS_resource_record_domain(inputs, outputs, state_machine):
        """
        Set the domain name for a DNS resource record.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS resource record.
                - The name of the variable containing the domain name to set.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the modified DNS resource record.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_resource_record = state_machine.get_variable_value(inputs[0])
        domain_name = state_machine.get_variable_value(inputs[1])

        dns_resource_record.rrname = domain_name

        state_machine.set_variable_value(outputs[0], dns_resource_record)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_DNS_resource_record_value(inputs, outputs, state_machine):
        """
        Set the value for a DNS resource record.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS resource record.
                - The name of the variable containing the value to set.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the modified DNS resource record.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_resource_record = state_machine.get_variable_value(inputs[0])
        value = state_machine.get_variable_value(inputs[1])

        dns_resource_record.rdata = value

        state_machine.set_variable_value(outputs[0], dns_resource_record)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def set_DNS_resource_record_type(inputs, outputs, state_machine):
        """
        Set the record type for a DNS resource record.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS resource record.
                - The name of the variable containing the record type to set.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the modified DNS resource record.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_resource_record = state_machine.get_variable_value(inputs[0])
        record_type = state_machine.get_variable_value(inputs[1])

        dns_resource_record.type = record_type

        state_machine.set_variable_value(outputs[0], dns_resource_record)

    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def add_DNS_response_to_answer(inputs, outputs, state_machine):
        """
        Add a DNS response packet to the answer section of a DNS packet.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS packet.
                - The name of the variable containing the DNS response packet to be added to the answer section.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        response_packet = state_machine.get_variable_value(inputs[1])

        dns_packet['DNS'].an = response_packet

        state_machine.set_variable_value(outputs[0], dns_packet)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def get_DNS_answer_from_DNS_packet(inputs, outputs, state_machine):
        """
        Get the DNS answer field from a DNS packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the DNS answer.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_answer = dns_packet.an

        state_machine.set_variable_value(outputs[0], dns_answer)


    @staticmethod
    @parsing_decorator(input_args=2, output_args=1)
    def add_DNS_response_to_additional(inputs, outputs, state_machine):
        """
        Add a DNS response packet to the additional answer section of a DNS packet.

        Number of input arguments: 2

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains two mandatory input arguments:
                - The name of the variable containing the DNS packet.
                - The name of the variable containing the DNS response packet to be added to the additional answer section.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is
                the name of the variable to store the modified DNS packet.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        additional_response = state_machine.get_variable_value(inputs[1])

        dns_packet['DNS'].ar = additional_response

        state_machine.set_variable_value(outputs[0], dns_packet)

    @staticmethod
    @parsing_decorator(input_args=1, output_args=1)
    def get_DNS_additional_answer_from_DNS_packet(inputs, outputs, state_machine):
        """
        Get the DNS additional answer field from a DNS packet.

        Number of input arguments: 1

        Number of output arguments: 1

        Optional input arguments: No

        Optional output arguments: No

        Args:
            inputs (List[str]): The list of input variable names. It contains one mandatory input argument, which is the name of the variable containing the DNS packet.

            outputs (List[str]): The list of output variable names. It contains one mandatory output argument, which is the name of the variable to store the DNS additional answer.

            state_machine: The state machine object.

        Returns:
            None
        """
        dns_packet = state_machine.get_variable_value(inputs[0])
        dns_additional_answer = dns_packet.ar

        state_machine.set_variable_value(outputs[0], dns_additional_answer)